@@ -22,20 +22,24 @@
 
 #include <X11/Xlib.h>
 #include <X11/extensions/Xrender.h>
+#include <stdbool.h>
 #include <stdio.h>
+#include <stdlib.h>
 
 #if HAVE_ERR_H
 # include <err.h>
 #else
 # include <stdarg.h>
 # include <stdlib.h>
-static inline void errx(int eval, const char *fmt, ...) {
+__attribute__((format(printf, 2, 3)))
+static inline void errx(int eval, const char *fmt, ...)
+{
     va_list args;
 
     va_start(args, fmt);
     fprintf(stderr, "Fatal Error: ");
-    fprintf(stderr, fmt, args);
-    fprintf(stderr, "\n", args);
+    vfprintf(stderr, fmt, args);
+    fprintf(stderr, "\n");
     va_end(args);
     exit(eval);
 }
@@ -43,11 +47,7 @@ static inline void errx(int eval, const char *fmt, ...) {
 
 #define min(a, b) (a < b ? a : b)
 #define max(a, b) (a > b ? a : b)
-
-#ifndef TRUE
-#define TRUE 1
-#define FALSE 0
-#endif
+#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))
 
 typedef struct _color4d
 {
@@ -65,9 +65,22 @@ typedef struct _picture_info {
 struct op_info {
 	int op;
 	const char *name;
-	Bool disabled;
+	bool disabled;
 };
 
+struct rendercheck_test_result {
+	int tests;
+	int passed;
+};
+
+static inline void
+record_result(struct rendercheck_test_result *result, bool success)
+{
+	result->tests++;
+	if (success)
+		result->passed++;
+}
+
 #define TEST_FILL		0x0001
 #define TEST_DSTCOORDS		0x0002
 #define TEST_SRCCOORDS		0x0004
@@ -81,13 +94,65 @@ struct op_info {
 #define TEST_REPEAT	  	0x0400
 #define TEST_TRIANGLES  	0x0800
 #define TEST_BUG7366		0x1000
-#define TEST_GTK_ARGB_XBGR	0x2000
-#define TEST_LIBREOFFICE_XRGB	0x4000
+#define TEST_gtk_argb_xbgr	0x2000
+#define TEST_libreoffice_xrgb	0x4000
+#define TEST_shmblend		0x8000
+
+struct rendercheck_test {
+	int bit;
+	const char *arg_name;
+	const char *long_name;
+	struct rendercheck_test_result (*func)(Display *dpy);
+};
+
+#ifdef __MACH__
+#define SECTION(X) section("__DATA,__" X )
+#define SECTION_START(X) __asm("section$start$__DATA$__" X)
+#define SECTION_END(X) __asm("section$end$__DATA$__" X)
+#else
+#define SECTION(X) section(X)
+#define SECTION_START(X)
+#define SECTION_END(X)
+#endif
+
+#ifdef __clang__
+#define ATTRIBUTE_NO_ASAN __attribute__((no_sanitize("address")))
+#else
+#define ATTRIBUTE_NO_ASAN /**/
+#endif
+
+#define DECLARE_RENDERCHECK_TEST(name)             \
+	const struct rendercheck_test test_desc_##name \
+	ATTRIBUTE_NO_ASAN \
+	__attribute__ ((used, SECTION ("test_section")))
+
+#define DECLARE_RENDERCHECK_ARG_TEST(arg_name_, long_name_, func_)		\
+	DECLARE_RENDERCHECK_TEST(arg_name_) = {				\
+		.bit = TEST_##arg_name_,				\
+		.arg_name = #arg_name_,					\
+		.long_name = long_name_,				\
+		.func = func_,						\
+	}
+
+/* Storage that will point at the start and end of the ELF/MACH-O section for test
+ * structs.  These are automatically set up by the linker when placing things
+ * in their sections.
+ */
+extern struct rendercheck_test __start_test_section SECTION_START("test_section");
+extern struct rendercheck_test __stop_test_section SECTION_END("test_section");
+
+struct render_format {
+	XRenderPictFormat *format;
+	char *name;
+};
+
+extern struct render_format *formats;
+extern int nformats;
 
 extern int pixmap_move_iter;
 extern int win_width, win_height;
 extern struct op_info ops[];
-extern Bool is_verbose, minimalrendering;
+extern bool is_verbose, minimalrendering;
 extern color4d colors[];
 extern int enabled_tests;
 extern int format_whitelist_len;
@@ -147,7 +212,7 @@ void
 argb_fill(Display *dpy, picture_info *p, int x, int y, int w, int h, float a,
     float r, float g, float b);
 
-Bool
+bool
 do_tests(Display *dpy, picture_info *win);
 
 void
@@ -161,69 +226,68 @@ do_composite(int op,
 	     const color4d *mask,
 	     const color4d *dst,
 	     color4d *result,
-	     Bool componentAlpha);
+	     bool componentAlpha);
 
 /* The tests */
-Bool
+bool
 blend_test(Display *dpy, picture_info *win, picture_info *dst,
 	   const int *op, int num_op,
 	   const picture_info **src_color, int num_src,
 	   const picture_info **dst_color, int num_dst);
 
-Bool
+bool
 composite_test(Display *dpy, picture_info *win, picture_info *dst,
 	       const int *op, int num_op,
 	       const picture_info **src_color, int num_src,
 	       const picture_info **mask_color, int num_mask,
 	       const picture_info **dst_color, int num_dst,
-	       Bool componentAlpha);
+	       bool componentAlpha);
 
-Bool
+bool
 dstcoords_test(Display *dpy, picture_info *win, int op, picture_info *dst,
     picture_info *bg, picture_info *fg);
 
-Bool
+bool
 fill_test(Display *dpy, picture_info *win, picture_info *src);
 
-Bool
+bool
 srccoords_test(Display *dpy, picture_info *win, picture_info *white,
-    Bool test_mask);
+    bool test_mask);
 
-Bool
+bool
 trans_coords_test(Display *dpy, picture_info *win, picture_info *white,
-    Bool test_mask);
+    bool test_mask);
 
-Bool
+bool
 trans_srccoords_test_2(Display *dpy, picture_info *win, picture_info *white,
-    Bool test_mask);
+    bool test_mask);
 
-Bool render_to_gradient_test(Display *dpy, picture_info *src);
+bool render_to_gradient_test(Display *dpy, picture_info *src);
 
-Bool linear_gradient_test(Display *dpy, picture_info *win,
+bool linear_gradient_test(Display *dpy, picture_info *win,
                           picture_info *dst, int op, picture_info *dst_color);
 
-Bool
+bool
 repeat_test(Display *dpy, picture_info *win, picture_info *dst, int op,
     picture_info *dst_color, picture_info *c1, picture_info *c2,
-    Bool test_mask);
+    bool test_mask);
 
-Bool
+bool
 triangles_test(Display *dpy, picture_info *win, picture_info *dst, int op,
     picture_info *src_color, picture_info *dst_color);
 
-Bool
+bool
 tristrip_test(Display *dpy, picture_info *win, picture_info *dst, int op,
     picture_info *src_color, picture_info *dst_color);
 
-Bool
+bool
 trifan_test(Display *dpy, picture_info *win, picture_info *dst, int op,
     picture_info *src_color, picture_info *dst_color);
 
-Bool
+bool
 bug7366_test(Display *dpy);
 
-Bool
-gtk_argb_xbgr_test(Display *dpy);
-
-Bool
-libreoffice_xrgb_test(Display *dpy, Bool invert);
+#define for_each_test(test)						\
+	for (struct rendercheck_test *test = &__start_test_section;	\
+	     test < &__stop_test_section;				\
+	     test++)
