@@ -27,8 +27,8 @@
 
 #include "rendercheck.h"
 
-static XRenderPictFormat **format_list;
-static int nformats;
+struct render_format *formats;
+int nformats;
 static int argb32index;
 
 /* Note: changing the order of these colors may disrupt tests that depend on
@@ -40,7 +40,10 @@ color4d colors[] = {
 	{1.0, 0, 0, 1.0},
 	{0, 1.0, 0, 1.0},
 	{0, 0, 1.0, 1.0},
+	{0.5, 0, 0, 0.25},
 	{0.5, 0, 0, .5},
+	{0.0, .5, 1.0, .5},
+	{0.0, .5, 1.0, 0}
 };
 
 /* Convenience pointers to 1x1 repeating colors */
@@ -113,16 +116,16 @@ color_correct(picture_info *pi, color4d *color)
 }
 
 void
-get_pixel(Display *dpy, picture_info *pi, int x, int y, color4d *color)
+get_pixel_from_image(XImage *image,
+		     const picture_info *pi,
+		     int x, int y,
+		     color4d *color)
 {
-	XImage *image;
 	unsigned long val;
 	unsigned long rm, gm, bm, am;
 	XRenderDirectFormat *layout = &pi->format->direct;
 
-	image = XGetImage(dpy, pi->d, x, y, 1, 1, 0xffffffff, ZPixmap);
-
-	val = XGetPixel(image, 0, 0);
+	val = XGetPixel(image, x, y);
 
 	rm = (unsigned long)layout->redMask << layout->red;
 	gm = (unsigned long)layout->greenMask << layout->green;
@@ -141,46 +144,78 @@ get_pixel(Display *dpy, picture_info *pi, int x, int y, color4d *color)
 		color->g = 0.0;
 		color->b = 0.0;
 	}
+}
+
+void
+get_pixel(Display *dpy,
+	  const picture_info *pi,
+	  int x, int y,
+	  color4d *color)
+{
+	XImage *image;
+
+	image = XGetImage(dpy, pi->d, x, y, 1, 1, 0xffffffff, ZPixmap);
+	get_pixel_from_image(image, pi, 0, 0, color);
 	XDestroyImage(image);
 }
 
-int
-eval_diff(char *name, color4d *expected, color4d *test, int x, int y,
-    Bool verbose)
+void
+accuracy(XRenderDirectFormat *result,
+	 const XRenderDirectFormat *a,
+	 const XRenderDirectFormat *b)
 {
-	double rscale, gscale, bscale, ascale;
-	double rdiff, gdiff, bdiff, adiff, diff;
+    result->redMask = min(a->redMask, b->redMask);
+    result->greenMask = min(a->greenMask, b->greenMask);
+    result->blueMask = min(a->blueMask, b->blueMask);
+    result->alphaMask = min(a->alphaMask, b->alphaMask);
+}
 
-	/* XXX: Need to be provided mask shifts so we can produce useful error
-	 * values.
-	 */
-	rscale = 1.0 * (1 << 5);
-	gscale = 1.0 * (1 << 6);
-	bscale = 1.0 * (1 << 5);
-	ascale = 1.0 * 32;
-	rdiff = fabs(test->r - expected->r) * rscale;
-	bdiff = fabs(test->g - expected->g) * gscale;
-	gdiff = fabs(test->b - expected->b) * bscale;
-	adiff = fabs(test->a - expected->a) * ascale;
-	/*rdiff = log2(1.0 + rdiff);
+double
+eval_diff(const XRenderDirectFormat *format,
+	  const color4d *expected,
+	  const color4d *test)
+{
+	double rdiff, gdiff, bdiff, adiff;
+
+	rdiff = fabs(test->r - expected->r) * format->redMask;
+	bdiff = fabs(test->g - expected->g) * format->greenMask;
+	gdiff = fabs(test->b - expected->b) * format->blueMask;
+	adiff = fabs(test->a - expected->a) * format->alphaMask;
+
+#if 0
+	rdiff = log2(1.0 + rdiff);
 	gdiff = log2(1.0 + gdiff);
 	bdiff = log2(1.0 + bdiff);
-	adiff = log2(1.0 + adiff);*/
-	diff = max(max(max(rdiff, gdiff), bdiff), adiff);
-	if (diff > 3.0) {
-		printf("%s test error of %.4f at (%d, %d) --\n"
-		    "           R    G    B    A\n"
-		    "got:       %.2f %.2f %.2f %.2f\n"
-		    "expected:  %.2f %.2f %.2f %.2f\n", name, diff, x, y,
-		    test->r, test->g, test->b, test->a,
-		    expected->r, expected->g, expected->b, expected->a);
-		return FALSE;
-	} else if (verbose) {
-		printf("%s test succeeded at (%d, %d) with %.4f: "
-		    "%.2f %.2f %.2f %.2f\n", name, x, y, diff,
-		    expected->r, expected->g, expected->b, expected->a);
-	}
-	return TRUE;
+	adiff = log2(1.0 + adiff);
+#endif
+
+	return max(max(max(rdiff, gdiff), bdiff), adiff);
+}
+
+void print_fail(const char *name,
+		const color4d *expected,
+		const color4d *test,
+		int x, int y,
+		double d)
+{
+    printf("%s test error of %.4f at (%d, %d) --\n"
+	   "           R     G     B     A\n"
+	   "got:       %.3f %.3f %.3f %.3f\n"
+	   "expected:  %.3f %.3f %.3f %.3f\n",
+	   name, d, x, y,
+	   test->r, test->g, test->b, test->a,
+	   expected->r, expected->g, expected->b, expected->a);
+}
+
+void print_pass(const char *name,
+		const color4d *expected,
+		int x, int y,
+		double d)
+{
+    printf("%s test succeeded at (%d, %d) with %.4f: "
+	   "%.2f %.2f %.2f %.2f\n",
+	   name, x, y, d,
+	   expected->r, expected->g, expected->b, expected->a);
 }
 
 void
@@ -197,8 +232,6 @@ argb_fill(Display *dpy, picture_info *p, int x, int y, int w, int h, float a,
 	XRenderFillRectangle(dpy, PictOpSrc, p->pict, &rendercolor, x, y, w, h);
 }
 
-#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))
-
 /* Create a set of direct format XRenderPictFormats for later use.  This lets
  * us get more formats than just the standard required set, and lets us attach
  * names to them.
@@ -208,71 +241,74 @@ create_formats_list(Display *dpy)
 {
     int i;
     int nformats_allocated = 5;
-    XRenderPictFormat templ;
+    XRenderPictFormat templ, *format;
 
     memset(&templ, 0, sizeof(templ));
     templ.type = PictTypeDirect;
 
-    format_list = malloc(sizeof(XRenderPictFormat *) * nformats_allocated);
-    if (format_list == NULL)
+    formats = calloc(sizeof(*formats), nformats_allocated);
+    if (formats == NULL)
 	errx(1, "malloc error");
     nformats = 0;
 
     argb32index = -1;
     for (i = 0; ; i++) {
-	char name[20];
 	int alphabits, redbits;
 
 	if (nformats + 1 == nformats_allocated) {
 	    nformats_allocated *= 2;
-	    format_list = realloc(format_list, sizeof(XRenderPictFormat *) *
-		nformats_allocated);
-	    if (format_list == NULL)
+	    formats = realloc(formats, sizeof(*formats) * nformats_allocated);
+	    if (formats == NULL)
 		errx(1, "realloc error");
 	}
 
-	format_list[nformats] = XRenderFindFormat(dpy, PictFormatType, &templ,
-	    i);
-	if (format_list[nformats] == NULL)
-	    break;
+	format = XRenderFindFormat(dpy, PictFormatType, &templ, i);
+	if (!format)
+		break;
 
-	alphabits = bit_count(format_list[nformats]->direct.alphaMask);
-	redbits = bit_count(format_list[nformats]->direct.redMask);
+	formats[nformats].format = format;
+
+	alphabits = bit_count(format->direct.alphaMask);
+	redbits = bit_count(format->direct.redMask);
 
 	/* Our testing code isn't all that hot, so don't bother trying at
 	 * the low depths yet.
 	 */
-	if ((redbits >= 1 && redbits <= 4) ||
-	    (alphabits >= 1 && alphabits <= 4))
+	if ((redbits >= 1 && redbits < 4) ||
+	    (alphabits >= 1 && alphabits < 4))
 	{
 	    continue;
 	}
 
-	describe_format(name, 20, format_list[nformats]);
+	describe_format(&formats[nformats].name, NULL, format);
 
-	if (format_whitelist_len != 0) {
-	    Bool ok = FALSE;
-	    int j;
-
-	    for (j = 0; j < format_whitelist_len; j++) {
-		if (strcmp(format_whitelist[j], name) == 0) {
-		    ok = TRUE;
-		    break;
-		}
-	    }
-	    if (!ok) {
-		printf("Ignoring server-supported format: %s\n", name);
-		continue;
-	    }
-	}
-
-	if (format_list[nformats] == XRenderFindStandardFormat(dpy,
-	    PictStandardARGB32))
+	if (format == XRenderFindStandardFormat(dpy, PictStandardARGB32))
 	{
 	    argb32index = nformats;
 	}
 
-	printf("Found server-supported format: %s\n", name);
+	if (format_whitelist_len != 0) {
+	    bool ok = false;
+	    int j;
+
+	    for (j = 0; j < format_whitelist_len; j++) {
+		if (strcmp(format_whitelist[j], formats[nformats].name) == 0) {
+		    ok = true;
+		    break;
+		}
+	    }
+	    if (!ok) {
+		printf("Ignoring server-supported format: %s\n",
+		       formats[nformats].name);
+		if (nformats != argb32index) {
+		    free(formats[nformats].name);
+		    formats[nformats].name = NULL;
+		}
+		continue;
+	    }
+	}
+
+	printf("Found server-supported format: %s\n", formats[nformats].name);
 
 	nformats++;
     }
@@ -281,14 +317,20 @@ create_formats_list(Display *dpy)
     }
 }
 
-Bool
+bool
 do_tests(Display *dpy, picture_info *win)
 {
-	int i, j, src, dst = 0, mask;
+	int i, j, src;
 	int num_dests;
 	picture_info *dests, *pictures_1x1, *pictures_10x10, picture_3x3, *pictures_solid;
 	int success_mask = 0, tests_passed = 0, tests_total = 0;
 	int num_tests;
+	int *test_ops;
+	const picture_info **test_src, **test_mask, **test_dst;
+	int num_test_ops = 0;
+	int num_test_src = 0;
+	int num_test_mask = 0;
+	int num_test_dst = 0;
 
 	create_formats_list(dpy);
 
@@ -298,16 +340,13 @@ do_tests(Display *dpy, picture_info *win)
 		errx(1, "malloc error");
 
 	for (i = 0; i < num_dests; i++) {
-		dests[i].format = format_list[i];
-		dests[i].d = XCreatePixmap(dpy, RootWindow(dpy, 0),
+		dests[i].format = formats[i].format;
+		dests[i].d = XCreatePixmap(dpy, DefaultRootWindow(dpy),
 		    win_width, win_height, dests[i].format->depth);
 		dests[i].pict = XRenderCreatePicture(dpy, dests[i].d,
 		    dests[i].format, 0, NULL);
 
-		dests[i].name = (char *)malloc(20);
-		if (dests[i].name == NULL)
-			errx(1, "malloc error");
-		describe_format(dests[i].name, 20, dests[i].format);
+		describe_format(&dests[i].name, NULL, dests[i].format);
 	}
 
 	pictures_1x1 = (picture_info *)malloc(num_colors * nformats *
@@ -320,20 +359,15 @@ do_tests(Display *dpy, picture_info *win)
 		color4d *c = &colors[i / nformats];
 
 		/* The standard PictFormat numbers go from 0 to 4 */
-		pictures_1x1[i].format = format_list[i % nformats];
-		pictures_1x1[i].d = XCreatePixmap(dpy, RootWindow(dpy, 0), 1,
-		    1, pictures_1x1[i].format->depth);
-		pa.repeat = TRUE;
+		pictures_1x1[i].format = formats[i % nformats].format;
+		pictures_1x1[i].d = XCreatePixmap(dpy, DefaultRootWindow(dpy),
+		    1, 1, pictures_1x1[i].format->depth);
+		pa.repeat = true;
 		pictures_1x1[i].pict = XRenderCreatePicture(dpy,
 		    pictures_1x1[i].d, pictures_1x1[i].format, CPRepeat, &pa);
 
-		pictures_1x1[i].name = (char *)malloc(20);
-		if (pictures_1x1[i].name == NULL)
-			errx(1, "malloc error");
-		sprintf(pictures_1x1[i].name, "1x1R ");
-		describe_format(pictures_1x1[i].name +
-		    strlen(pictures_1x1[i].name), 20 -
-		    strlen(pictures_1x1[i].name), pictures_1x1[i].format);
+		describe_format(&pictures_1x1[i].name, "1x1R ",
+		    pictures_1x1[i].format);
 
 		argb_fill(dpy, &pictures_1x1[i], 0, 0, 1, 1,
 		    c->a, c->r, c->g, c->b);
@@ -352,24 +386,17 @@ do_tests(Display *dpy, picture_info *win)
 		errx(1, "malloc error");
 
 	for (i = 0; i < num_colors * nformats; i++) {
-		XRenderPictureAttributes pa;
 		color4d *c = &colors[i / nformats];
 
 		/* The standard PictFormat numbers go from 0 to 4 */
-		pictures_10x10[i].format = format_list[i % nformats];
-		pictures_10x10[i].d = XCreatePixmap(dpy, RootWindow(dpy, 0), 10,
-		    10, pictures_10x10[i].format->depth);
-		pa.repeat = TRUE;
+		pictures_10x10[i].format = formats[i % nformats].format;
+		pictures_10x10[i].d = XCreatePixmap(dpy, DefaultRootWindow(dpy),
+		    10, 10, pictures_10x10[i].format->depth);
 		pictures_10x10[i].pict = XRenderCreatePicture(dpy,
 		    pictures_10x10[i].d, pictures_10x10[i].format, 0, NULL);
 
-		pictures_10x10[i].name = (char *)malloc(20);
-		if (pictures_10x10[i].name == NULL)
-			errx(1, "malloc error");
-		sprintf(pictures_10x10[i].name, "10x10 ");
-		describe_format(pictures_10x10[i].name +
-		    strlen(pictures_10x10[i].name), 20 -
-		    strlen(pictures_10x10[i].name), pictures_10x10[i].format);
+		describe_format(&pictures_10x10[i].name, "10x10 ",
+		    pictures_10x10[i].format);
 
 		argb_fill(dpy, &pictures_10x10[i], 0, 0, 10, 10,
 		    c->a, c->r, c->g, c->b);
@@ -378,11 +405,11 @@ do_tests(Display *dpy, picture_info *win)
 		color_correct(&pictures_10x10[i], &pictures_10x10[i].color);
 	}
 
-	picture_3x3.d = XCreatePixmap(dpy, RootWindow(dpy, 0), 3, 3, 32);
+	picture_3x3.d = XCreatePixmap(dpy, DefaultRootWindow(dpy), 3, 3, 32);
 	picture_3x3.format = XRenderFindStandardFormat(dpy, PictStandardARGB32);
 	picture_3x3.pict = XRenderCreatePicture(dpy, picture_3x3.d,
 	    picture_3x3.format, 0, NULL);
-	picture_3x3.name = "3x3 sample picture";
+	picture_3x3.name = (char *)"3x3 sample picture";
 	for (i = 0; i < 9; i++) {
 		int x = i % 3;
 		int y = i / 3;
@@ -400,8 +427,8 @@ do_tests(Display *dpy, picture_info *win)
             c.green = (int)(colors[i].g*65535);
             c.blue = (int)(colors[i].b*65535);
             pictures_solid[i].pict = XRenderCreateSolidFill(dpy, &c);
-            pictures_solid[i].format = format_list[argb32index];
-            pictures_solid[i].name = "Solid";
+            pictures_solid[i].format = formats[argb32index].format;
+            pictures_solid[i].name = (char *)"Solid";
         }
 
 #define RECORD_RESULTS()					\
@@ -414,8 +441,46 @@ do {								\
 
 	num_tests = num_colors * nformats;
 
+	test_ops = malloc(sizeof(int)*num_ops);
+	test_src = malloc(sizeof(picture_info*)*(2*num_tests+num_colors));
+	test_mask = malloc(sizeof(picture_info*)*2*num_tests);
+	test_dst = malloc(sizeof(picture_info*)*(num_tests+num_colors));
+
+	for (i = 0; i < num_ops; i++) {
+	    if (ops[i].disabled)
+		continue;
+
+	    test_ops[num_test_ops++] = i;
+	}
+
+	for (i = 0; i < num_tests; i++) {
+	    test_src[num_test_src++] = &pictures_1x1[i];
+	    test_src[num_test_src++] = &pictures_10x10[i];
+	    test_mask[num_test_mask++] = &pictures_1x1[i];
+	    test_mask[num_test_mask++] = &pictures_10x10[i];
+	    test_dst[num_test_dst++] = &pictures_1x1[i];
+	}
+	for (i = 0; i < num_colors; i++) {
+	    test_src[num_test_src++] = &pictures_solid[i];
+	    test_dst[num_test_dst++] = &pictures_solid[i];
+	}
+
+	for_each_test(test) {
+		struct rendercheck_test_result result;
+
+		if (!(enabled_tests & test->bit))
+			continue;
+
+		result = test->func(dpy);
+		tests_total += result.tests;
+		tests_passed += result.passed;
+
+		if (result.tests == result.passed)
+			success_mask |= test->bit;
+	}
+
 	if (enabled_tests & TEST_FILL) {
-		Bool ok, group_ok = TRUE;
+		bool ok, group_ok = true;
 
 		printf("Beginning testing of filling of 1x1R pictures\n");
 		for (i = 0; i < num_tests; i++) {
@@ -433,7 +498,7 @@ do {								\
 	}
 
 	if (enabled_tests & TEST_DSTCOORDS) {
-		Bool ok, group_ok = TRUE;
+		bool ok, group_ok = true;
 
 		printf("Beginning dest coords test\n");
 		for (i = 0; i < 2; i++) {
@@ -447,48 +512,48 @@ do {								\
 	}
 
 	if (enabled_tests & TEST_SRCCOORDS) {
-		Bool ok, group_ok = TRUE;
+		bool ok, group_ok = true;
 
 		printf("Beginning src coords test\n");
-		ok = srccoords_test(dpy, win, argb32white, FALSE);
+		ok = srccoords_test(dpy, win, argb32white, false);
 		RECORD_RESULTS();
 		if (group_ok)
 			success_mask |= TEST_SRCCOORDS;
 	}
 
 	if (enabled_tests & TEST_MASKCOORDS) {
-		Bool ok, group_ok = TRUE;
+		bool ok, group_ok = true;
 
 		printf("Beginning mask coords test\n");
-		ok = srccoords_test(dpy, win, argb32white, TRUE);
+		ok = srccoords_test(dpy, win, argb32white, true);
 		RECORD_RESULTS();
 		if (group_ok)
 			success_mask |= TEST_MASKCOORDS;
 	}
 
 	if (enabled_tests & TEST_TSRCCOORDS) {
-		Bool ok, group_ok = TRUE;
+		bool ok, group_ok = true;
 
 		printf("Beginning transformed src coords test\n");
-		ok = trans_coords_test(dpy, win, argb32white, FALSE);
+		ok = trans_coords_test(dpy, win, argb32white, false);
 		RECORD_RESULTS();
 
 		printf("Beginning transformed src coords test 2\n");
-		ok = trans_srccoords_test_2(dpy, win, argb32white, FALSE);
+		ok = trans_srccoords_test_2(dpy, win, argb32white, false);
 		RECORD_RESULTS();
 		if (group_ok)
 			success_mask |= TEST_TSRCCOORDS;
 	}
 
 	if (enabled_tests & TEST_TMASKCOORDS) {
-		Bool ok, group_ok = TRUE;
+		bool ok, group_ok = true;
 
 		printf("Beginning transformed mask coords test\n");
-		ok = trans_coords_test(dpy, win, argb32white, TRUE);
+		ok = trans_coords_test(dpy, win, argb32white, true);
 		RECORD_RESULTS();
 
 		printf("Beginning transformed mask coords test 2\n");
-		ok = trans_srccoords_test_2(dpy, win, argb32white, TRUE);
+		ok = trans_srccoords_test_2(dpy, win, argb32white, true);
 		RECORD_RESULTS();
 
 		if (group_ok)
@@ -496,146 +561,80 @@ do {								\
 	}
 
 	if (enabled_tests & TEST_BLEND) {
-		Bool ok, group_ok = TRUE;
+		bool ok, group_ok = true;
 
-		for (i = 0; i < num_ops; i++) {
-		    if (ops[i].disabled)
-			continue;
+		for (j = 0; j <= num_dests; j++) {
+		    picture_info *pi;
 
-		    for (j = 0; j <= num_dests; j++) {
-			picture_info *pi;
+		    if (j != num_dests)
+			pi = &dests[j];
+		    else
+			pi = win;
 
-			if (j != num_dests)
-				pi = &dests[j];
-			else
-				pi = win;
-			printf("Beginning %s blend test on %s\n", ops[i].name,
-			    pi->name);
+		    printf("Beginning blend test on %s\n", pi->name);
 
-			for (src = 0; src < num_tests; src++) {
-				for (dst = 0; dst < num_tests; dst++) {
-					ok = blend_test(dpy, win, pi, i,
-					    &pictures_1x1[src],
-					    &pictures_1x1[dst]);
-					RECORD_RESULTS();
-					ok = blend_test(dpy, win, pi, i,
-					    &pictures_10x10[src],
-					    &pictures_1x1[dst]);
-					RECORD_RESULTS();
-				}
-			}
-                        for (src = 0; src < num_colors; src++) {
-                            for (dst = 0; dst < num_colors; dst++) {
-                                ok = blend_test(dpy, win, pi, i, &pictures_solid[src], &pictures_1x1[dst]);
-				RECORD_RESULTS();
-                            }
-                        }
-		    }
+		    ok = blend_test(dpy, win, pi,
+				    test_ops, num_test_ops,
+				    test_src, num_test_src,
+				    test_dst, num_test_dst);
+		    RECORD_RESULTS();
 		}
 		if (group_ok)
 			success_mask |= TEST_BLEND;
 	}
 
 	if (enabled_tests & TEST_COMPOSITE) {
-		Bool ok, group_ok = TRUE;
+		bool ok, group_ok = true;
 
-		for (i = 0; i < num_ops; i++) {
-		    if (ops[i].disabled)
-			continue;
+		for (j = 0; j <= num_dests; j++) {
+		    picture_info *pi;
 
-		    for (j = 0; j <= num_dests; j++) {
-			picture_info *pi;
+		    if (j != num_dests)
+			pi = &dests[j];
+		    else
+			pi = win;
 
-			if (j != num_dests)
-				pi = &dests[j];
-			else
-				pi = win;
-			printf("Beginning %s composite mask test on %s\n",
-			    ops[i].name, pi->name);
+		    printf("Beginning composite mask test on %s\n", pi->name);
 
-			for (src = 0; src < num_tests; src++) {
-			    for (mask = 0; mask < num_tests; mask++) {
-				for (dst = 0; dst < num_tests; dst++) {
-					ok = composite_test(dpy, win, pi, i,
-					    &pictures_10x10[src],
-					    &pictures_10x10[mask],
-					    &pictures_1x1[dst], FALSE, TRUE);
-					RECORD_RESULTS();
-					ok = composite_test(dpy, win, pi, i,
-					    &pictures_1x1[src],
-					    &pictures_10x10[mask],
-					    &pictures_1x1[dst], FALSE, TRUE);
-					RECORD_RESULTS();
-					ok = composite_test(dpy, win, pi, i,
-					    &pictures_10x10[src],
-					    &pictures_1x1[mask],
-					    &pictures_1x1[dst], FALSE, TRUE);
-					RECORD_RESULTS();
-					ok = composite_test(dpy, win, pi, i,
-					    &pictures_1x1[src],
-					    &pictures_1x1[mask],
-					    &pictures_1x1[dst], FALSE, TRUE);
-					RECORD_RESULTS();
-				}
-			    }
-			}
-		    }
+		    ok = composite_test(dpy, win, pi,
+					test_ops, num_test_ops,
+					test_src, num_test_src,
+					test_mask, num_test_mask,
+					test_dst, num_test_dst,
+					false);
+		    RECORD_RESULTS();
 		}
 		if (group_ok)
 			success_mask |= TEST_COMPOSITE;
 	}
 
 	if (enabled_tests & TEST_CACOMPOSITE) {
-		Bool ok, group_ok = TRUE;
+		bool ok, group_ok = true;
 
-		for (i = 0; i < num_ops; i++) {
-		    if (ops[i].disabled)
-			continue;
+		for (j = 0; j <= num_dests; j++) {
+		    picture_info *pi;
 
-		    for (j = 0; j <= num_dests; j++) {
-			picture_info *pi;
+		    if (j != num_dests)
+			pi = &dests[j];
+		    else
+			pi = win;
 
-			if (j != num_dests)
-				pi = &dests[j];
-			else
-				pi = win;
-			printf("Beginning %s composite CA mask test on %s\n",
-			    ops[i].name, pi->name);
+		    printf("Beginning composite CA mask test on %s\n", pi->name);
 
-			for (src = 0; src < num_tests; src++) {
-			    for (mask = 0; mask < num_tests; mask++) {
-				for (dst = 0; dst < num_tests; dst++) {
-					ok = composite_test(dpy, win, pi, i,
-					    &pictures_10x10[src],
-					    &pictures_10x10[mask],
-					    &pictures_1x1[dst], TRUE, TRUE);
-					RECORD_RESULTS();
-					ok = composite_test(dpy, win, pi, i,
-					    &pictures_1x1[src],
-					    &pictures_10x10[mask],
-					    &pictures_1x1[dst], TRUE, TRUE);
-					RECORD_RESULTS();
-					ok = composite_test(dpy, win, pi, i,
-					    &pictures_10x10[src],
-					    &pictures_1x1[mask],
-					    &pictures_1x1[dst], TRUE, TRUE);
-					RECORD_RESULTS();
-					ok = composite_test(dpy, win, pi, i,
-					    &pictures_1x1[src],
-					    &pictures_1x1[mask],
-					    &pictures_1x1[dst], TRUE, TRUE);
-					RECORD_RESULTS();
-				}
-			    }
-			}
-		    }
+		    ok = composite_test(dpy, win, pi,
+					test_ops, num_test_ops,
+					test_src, num_test_src,
+					test_mask, num_test_mask,
+					test_dst, num_test_dst,
+					true);
+		    RECORD_RESULTS();
 		}
 		if (group_ok)
 			success_mask |= TEST_CACOMPOSITE;
 	}
 
         if (enabled_tests & TEST_GRADIENTS) {
-	    Bool ok, group_ok = TRUE;
+	    bool ok, group_ok = true;
 
 	    printf("Beginning render to linear gradient test\n");
 	    ok = render_to_gradient_test(dpy, &pictures_1x1[0]);
@@ -667,7 +666,7 @@ do {								\
         }
 
         if (enabled_tests & TEST_REPEAT) {
-	    Bool ok, group_ok = TRUE;
+	    bool ok, group_ok = true;
 
             for (i = 0; i < num_ops; i++) {
 		if (ops[i].disabled)
@@ -686,7 +685,7 @@ do {								\
 		     * consisting of colors 1 and 2 (r, g).
 		     */
 		    ok = repeat_test(dpy, win, pi, i, argb32white, argb32red,
-		        argb32green, FALSE);
+		        argb32green, false);
 		    RECORD_RESULTS();
 
                     printf("Beginning %s mask repeat test on %s\n",
@@ -695,7 +694,7 @@ do {								\
 		     * repeating mask consisting of colors 1 and 2 (r, g).
 		     */
 		    ok = repeat_test(dpy, win, pi, i, argb32white, argb32red,
-		        argb32green, TRUE);
+		        argb32green, true);
 		    RECORD_RESULTS();
                 }
             }
@@ -704,7 +703,7 @@ do {								\
         }
 
 	if (enabled_tests & TEST_TRIANGLES) {
-	    Bool ok, group_ok = TRUE;
+	    bool ok, group_ok = true;
 
 	    for (i = 0; i < num_ops; i++) {
 		if (ops[i].disabled)
@@ -742,7 +741,7 @@ do {								\
 	}
 
         if (enabled_tests & TEST_BUG7366) {
-	    Bool ok, group_ok = TRUE;
+	    bool ok, group_ok = true;
 
 	    ok = bug7366_test(dpy);
 	    RECORD_RESULTS();
@@ -751,7 +750,32 @@ do {								\
 		success_mask |= TEST_BUG7366;
 	}
 
+	for (i = 0; i < num_colors * nformats; i++) {
+	    free(pictures_1x1[i].name);
+	    free(pictures_10x10[i].name);
+	}
+	free(pictures_1x1);
+	free(pictures_10x10);
+	free(pictures_solid);
+
+	for (i = 0; i < num_dests; i++) {
+	    free(dests[i].name);
+	}
+	free(dests);
+
+	for (i = 0; i < nformats; i++) {
+	    free(formats[i].name);
+	}
+	free(formats);
+
+	free(test_ops);
+	free(test_src);
+	free(test_mask);
+	free(test_dst);
+
 	printf("%d tests passed of %d total\n", tests_passed, tests_total);
+	printf("Successful Groups:\n");
+	print_tests(stdout, success_mask);
 
 	return tests_passed == tests_total;
 }
